Unit AdvMM;

 {

  System 2001 version 7
  --------------------------------------------------------------------------
  (C) 1997-2001, Rozsnyo electronics and software


    Advanced Memory Manager


    From: 2001-05-10
    By:   Daniel Rozsnyo


    Versioning:

      1.00 until 2001-05-10  Basic version
      2.00 from  2001-11-14  XMScap ariable added - block mode
                             ( you need this under true DOS )

 }

 Interface

   Type

     PXMSpart      = ^TXMSpart;
     TXMSpart      = record
                       Handle    : word;       { xms handle of block }
                       Size      : word;       { size of this block  }
                       Next      : PXMSpart;   { next block          }
                     end;

     PammPart      = ^TAMMpart;
     TammPart      = record
                       prev      : PAMMpart;
                       next      : PAMMpart;
                       offset    : word;
                       xmsblock  : PXMSpart;
                       size      : word;
                       isFree    : boolean;
                     end;

     TAMM          = object
                       XMSroot   : PXMSpart; { list of xms blocks        }

                       MaxMemory : longint;  { size of all memory        }
                       FreeMem   : longint;  { size of free memory       }
                       AllocMem  : longint;  { size of allocated memory  }

                       AllocRoot : Pammpart; { list of blocks            }

                       constructor Init;
                       procedure   AllocateXMS;
                       procedure   AllocInit;

                       procedure   SeekXMS(var block:Pxmspart;
                                           var o:word; size:word);

                       function    isValid(block:PammPart):boolean;

                       function    Alloc(ms:word):pointer;
                       function    DeAlloc(var block:PammPart):boolean;
                       function    Resize(var block:PammPart;
                                              diff:longint):boolean;

                       function    GetMaxSize(block:PammPart):longint;

                       procedure   Read(block:PammPart;ofs,count:longint;
                                                       buffer:pointer);
                       procedure   Write(block:PammPart;ofs,count:longint;
                                                        buffer:pointer);

                       procedure   LogDeAlloc;
                       procedure   FreeXMS;
                       destructor  Done;

                     end;

   Var

     XMS           : Tamm;

     XMScap        : byte;
                   {
                      XMSCAP = xms driver transfer capability


                        1 = you can transfer data of any size

                        2 = you can transfer only data, which have
                            the size multiple of 2 (words)

                        4 = you can transfer only data, which have
                            the size multiple of 4 (doublewords)


                   }

 Implementation

  uses XMScore;

  type ByteArray = array[0..255] of byte;

  var  CapBuff   : array[0..3] of byte;

  constructor Tamm.Init;
              begin
                {
                  prepare xms block list
                }
                XMSroot:=nil;
                MaxMemory:=0;
                {
                  if we have
                }
                if xmscore.xmsinstalled then AllocateXMS;
                {
                  prepare list of allocation blocks
                }
                Allocroot:=nil;
                {
                  create first empty block
                }
                AllocInit;
              end;

  procedure   Tamm.AllocateXMS;
              var current   : Pxmspart;
                  temp      : Pxmspart;
                  newhandle : word;
                  size      : word;
              begin
                {
                  deallocate old blocks
                }
                FreeXMS;
                {
                  prepare first block
                }
                current:=xmsroot;
                {
                  while exist free block...
                }
                size:=xmscore.xmsfreeblock;
                while size>0 do
                  begin
                    {
                      allocate the block
                    }
                    newhandle:=xmscore.xmsopen(size);
                    {
                      if the allocation was not successful...
                    }
                    if xmscore.xmserror<>0 then break;
                    {
                      allocate the memory for xmsroot
                    }
                    getmem(temp,sizeof(temp^));
                    {
                      if exist previous item, lint to its end
                    }
                    if current<>nil then current^.next:=temp
                                    else xmsroot:=temp;
                    {
                      switch the current item to new
                    }
                    current:=temp;
                    {
                      remember the size and the handle
                    }
                    current^.size:=size;
                    current^.handle:=newhandle;
                    current^.next:=nil;
                    {
                      add the memory size to counter
                    }
                    inc(maxmemory,size);
                    {
                      next block...
                    }
                    size:=xmscore.xmsfreeblock;
                  end;
              end;

  procedure   Tamm.AllocInit;
              begin
                {
                  deallocate allocation info blocks
                }
                LogDeAlloc;
                {
                  make one large block
                }
                system.getmem(AllocRoot,sizeof(allocroot^));
                {
                  fill values
                }
                with allocroot^ do
                     begin
                       prev    :=nil;
                       next    :=nil;
                       xmsblock:=xmsroot;
                       offset  :=0;
                       size    :=maxmemory;
                       isFree  :=true;
                     end;
                {
                  fast counter initialization
                }
                FreeMem:=maxmemory;
                AllocMem:=0;
              end;

  procedure   Tamm.LogDeAlloc;
              var temp : Pammpart;
              begin
                {
                  if is some memory allocated
                }
                while allocroot<>nil do
                      begin
                        {
                          remember the current item
                        }
                        temp:=allocroot;
                        {
                          move to next
                        }
                        allocroot:=allocroot^.next;
                        {
                          dealloc remembered (previous) item
                        }
                        system.freemem(temp,sizeof(temp^));
                      end;
                {
                  list is empty
                }
                allocroot:=nil;
              end;

  procedure   Tamm.FreeXMS;
              var current : Pxmspart;
                  temp    : Pxmspart;
              begin
                {
                  prepare first block
                }
                current:=xmsroot;
                {
                  if the block exist, do this...
                }
                if current<>nil then
                repeat
                  {
                    dealloc the block
                  }
                  xmscore.xmsclose(current^.handle);
                  {
                    remember the current item
                  }
                  temp:=current;
                  {
                    move to next
                  }
                  current:=current^.next;
                  {
                    deallocate remembered (previous) item
                  }
                  system.freemem(temp,sizeof(temp^));
                  {
                    do this while not end
                  }
                until current=nil;
                {
                  xms blocks removed
                }
                xmsRoot:=nil;
                MaxMemory:=0;
              end;

  destructor  Tamm.Done;
              begin
                {
                  deallocate allocation blocks
                }
                LogDeAlloc;
                {
                  deallocate xms memory
                }
                FreeXMS;
              end;

  procedure   Tamm.SeekXMS(var block:Pxmspart;var o:word; size:word);
              begin
                {
                  move the offset to wanted position
                }
                o:=o+size;
                {
                  if it is not in the current block....
                }
                while o>=block^.size do
                      begin
                        {
                          change block
                        }
                        o:=o-block^.size;
                        block:=block^.next;
                      end;
              end;

  function    Tamm.Alloc(ms:word):pointer;
              var CurrAB : Pammpart;
                  Temp   : Pammpart;
              begin
                {
                  We want to allocate ms KB of xms memory
                }
                {
                  Check the fragmented memory size
                }
                if ms>freemem then
                   begin
                     {
                       we know, that there is no way to alloc ms KB of mem
                     }
                     Alloc:=nil;
                     exit;
                   end;
                {
                  find free block
                }
                CurrAB:=AllocRoot;
                while currAB<>nil do
                      begin
                        {
                          block must be free and with enough free space
                        }
                        if (currAB^.isFree) and
                           (currAB^.size>=ms) then break;
                        {
                          next item...
                        }
                        currAB:=currAB^.next;
                      end;
                {
                  we found block or not - look to currAB variable
                }
                if CurrAB=nil then
                   begin
                     {
                       no allocation block with enough free space
                     }
                     Alloc:=nil;
                   end else
                   begin
                     {
                       we found allocation block with enough free space,
                       make Allocated block from free block
                     }
                     currAB^.isFree:=false;
                     {
                       return value;
                     }
                     Alloc:=currAB;
                     {
                       modify fast counters
                     }
                     inc(AllocMem,ms);
                     dec(Freemem,ms);
                     if currAB^.size<>ms then
                        begin
                          {
                            we found larger block - alloc new alloc. block
                          }
                          system.getmem(temp,sizeof(temp^));
                          {
                            set attributes
                          }
                          with temp^ do
                               begin
                                 isFree:=true;
                                 size:=currAB^.size-ms;
                                 prev:=currAB;
                                 next:=currAB^.next;
                                 xmsblock:=currAB^.xmsblock;
                                 offset:=currAB^.offset;
                                 SeekXMS(xmsblock,offset,ms);
                               end;
                          currAB^.size:=ms;
                          {
                            insert the item to the chain
                          }
                          if currAB^.next<> nil then
                             currAB^.next^.prev:=temp;
                          currAB^.next:=temp;
                          currAB^.isFree:=false;
                        end;
                   end;
              end;

  function    Tamm.isValid(block:PammPart):boolean;
              var curr : PammPart;
              begin
                {
                  start searching of block
                }
                curr:=allocroot;
                while curr<>nil do
                      begin
                        {
                          block found
                        }
                        if curr=block then break;
                        {
                          move to next item
                        }
                        curr:=curr^.next;
                      end;
                {
                  set the result
                }
                isValid:=curr=block;
              end;

  function    Tamm.DeAlloc(var block:PammPart):boolean;
              var curr : PammPart;
                  temp : PammPart;
              begin
                if not isValid(block) then
                   begin
                     {
                       we dont found the allocation block (=> exception)
                     }
                     deAlloc:=false;
                   end else
                   begin
                     curr:=block;
                     {
                       set return value -> indicates the successful dealloc
                     }
                     block:=nil;
                     {
                       deallocation of block - set isTrue
                     }
                     dealloc:=true;
                     curr^.isFree:=true;
                     {
                       update fast counters
                     }
                     inc(freemem, curr^.size);
                     dec(allocmem,curr^.size);
                     {
                       we look to the next block - exist and is free ?
                     }
                     if curr^.next<>nil then
                     if curr^.next^.isFree then
                        begin
                          {
                            join the current block with next
                          }
                          temp:=curr^.next;
                          inc(curr^.size,temp^.size);
                          {
                            re-link the blocks
                          }
                          curr^.next:=temp^.next;
                          if temp^.next<>nil then temp^.next^.prev:=curr;
                          {
                            dealloc the memory block
                          }
                          system.FreeMem(temp,sizeof(temp^));
                        end;
                     {
                       we look to the previous block - exist and is free ?
                     }
                     if curr^.prev<>nil then
                     if curr^.prev^.isFree then
                        begin
                          {
                            join the current block with previous
                          }
                          temp:=curr;
                          curr:=curr^.prev;
                          inc(curr^.size,temp^.size);
                          {
                            re-link the blocks
                          }
                          curr^.next:=temp^.next;
                          if temp^.next<>nil then temp^.next^.prev:=curr;
                          {
                            dealloc the memory block
                          }
                          system.FreeMem(temp,sizeof(temp^));
                        end;
                   end;
              end;

  function    Tamm.GetMaxSize(block:PammPart):longint;
              begin
                {
                  counts the maximal size of this block (for resizing)
                }
                GetMaxSize:=0;
                if isValid(block) then
                   if block^.next<>nil then
                   if block^.next^.isFree then
                      GetMaxSize:=block^.size+block^.next^.size;
              end;

  function    Tamm.Resize(var block:PammPart;diff:longint):boolean;
              var temp : PammPart;
              begin
                Resize:=false;
                if diff<>0 then
                if isValid(block) then
                   begin
                     if diff>0 then
                     if (block^.size+diff)<=GetMaxSize(block) then
                        begin
                          {
                            make this block larger
                          }
                          temp:=block^.next;
                          with temp^ do
                               begin
                                 dec(size,diff);
                                 SeekXMS(xmsblock,offset,diff);
                               end;
                          inc(block^.size,diff);
                          if temp^.size=0 then DeAlloc(temp);
                          dec(FreeMem,diff);
                          inc(AllocMem,diff);
                        end;
                     if diff<0 then
                     if (block^.size+diff)=0 then
                        {
                          if resized to 0
                        }
                        DeAlloc(block) else
                     if (block^.size+diff)>0 then
                        begin
                          diff:=-diff;
                          {
                            make this block smaller
                          }
                          dec(block^.size,diff);
                          temp:=block^.next;
                          with temp^ do
                               begin
                                 inc(size,diff);
                                 xmsblock:=block^.xmsblock;
                                 offset:=block^.offset;
                                 SeekXMS(xmsblock,offset,block^.size);
                               end;
                          inc(FreeMem,diff);
                          dec(AllocMem,diff);
                        end;
                   end;
              end;

  procedure   Tamm.Read(block:PammPart;ofs,count:longint;buffer:pointer);
              type ByteA  = array[0..1] of byte;
              var  realblock  : PxmsPart;
                   realoffset : longint;
                   realofs    : word;
                   pre,post   : longint;
                   main       : longint;
                   i          : byte;
              begin

                if ((ofs   mod xmscap)>0) or
                   ((count mod xmscap)>0) then
                   begin
                     {
                       the XMS driver dont support directly this move
                       we move data in 3 phases: pre,main,post

                          01234567 89abcdef 01234567 89abcdef
                         |        |        |        |        |
                         |  aaaaaa|bbbbbbbb|bbbbbbbb|ccccc   |
                         |        |        |        |        |
                          B1       B2       B3       B4       B5

                     }
                     pre:=xmscap-(ofs mod xmscap);    { A }
                     post:=(ofs+count) mod xmscap;    { C }
                     main:=count-pre-post;            { B }
                     {
                       PRE part
                     }
                     if pre>0 then
                        begin
                          read(block,ofs+pre-xmscap,xmscap,addr(capbuff));
                          for i:=0 to pre-1 do
                              ByteArray(buffer^)[i]:=capbuff[xmscap-pre+i];
                        end;
                     {
                       POST part
                     }
                     if post>0 then
                        begin
                          read(block,ofs+pre+main,xmscap,addr(capbuff));
                          for i:=0 to post-1 do
                              ByteArray(buffer^)[i+pre+main]:=capbuff[i];
                        end;
                     {
                       MAIN part
                     }
                     if main=0 then exit;
                     inc(ofs,pre);
                     count:=main;
                     buffer:=addr(ByteArray(buffer^)[pre]);
                   end;

                {
                  prepare position
                }
                realblock  := block^.xmsblock;
                realofs    := block^.offset;
                {
                  move the cursor by OFS/1024 Kilobytes
                }
                SeekXMS(realblock,realofs,ofs div 1024);
                {
                  add the rest of OFS to offset and convert it to Bytes
                }
                realoffset:=(longint(realofs)*1024) + (ofs mod 1024);
                {
                  now do the transfer (multiblock)
                }
                while count>0 do
                      begin
                        {
                          if the all bytes are in this block
                        }
                        if (realoffset+count)<=
                           (longint(realblock^.size)*1024) then
                           begin
                             with xmscore.xmsstruct do
                               begin
                                 length         := count;
                                 sourceh        := realblock^.handle;
                                 longint(source):= realoffset;
                                 desth          := 0;
                                 destination    := buffer;
                               end;
                             xmscore.xmsmove;
                             count:=0;
                           end else
                           begin
                             with xmscore.xmsstruct do
                               begin
                                 length         := (longint(realblock^.size)
                                                    *1024)-realoffset;
                                 sourceh        := realblock^.handle;
                                 longint(source):= realoffset;
                                 desth          := 0;
                                 destination    := buffer;
                                 dec(count,length);
                                 dec(realoffset,
                                     longint(realblock^.size)*1024-length);
                                 {
                                   move the pointer
                                 }
                                 buffer:=addr(ByteA(buffer^)[length]);
                               end;
                             xmscore.xmsmove;
                             {
                               select next block
                             }
                             realblock:=realblock^.next;
                           end;

                      end;

              end;

  procedure   Tamm.Write(block:PammPart;ofs,count:longint;buffer:pointer);
              type ByteA  = array[0..1] of byte;
              var  realblock  : PxmsPart;
                   realoffset : longint;
                   realofs    : word;
                   pre,post   : longint;
                   main       : longint;
                   i          : byte;
              begin

                if ((ofs   mod xmscap)>0) or
                   ((count mod xmscap)>0) then
                   begin
                     {
                       the XMS driver dont support directly this move
                       we move data in 3 phases: pre,main,post

                          01234567 89abcdef 01234567 89abcdef
                         |        |        |        |        |
                         |  aaaaaa|bbbbbbbb|bbbbbbbb|ccccc   |
                         |        |        |        |        |
                          B1       B2       B3       B4       B5

                     }
                     pre:=xmscap-(ofs mod xmscap);    { A }
                     post:=(ofs+count) mod xmscap;    { C }
                     main:=count-pre-post;            { B }
                     {
                       PRE part
                     }
                     if pre>0 then
                        begin
                          read(block,ofs+pre-xmscap,xmscap,addr(capbuff));
                          for i:=0 to pre-1 do
                              capbuff[xmscap-pre+i]:=ByteArray(buffer^)[i];
                          write(block,ofs+pre-xmscap,xmscap,addr(capbuff));
                        end;
                     {
                       POST part
                     }
                     if post>0 then
                        begin
                          read(block,ofs+pre+main,xmscap,addr(capbuff));
                          for i:=0 to post-1 do
                              capbuff[i]:=ByteArray(buffer^)[i+pre+main];
                          write(block,ofs+pre+main,xmscap,addr(capbuff));
                        end;
                     {
                       MAIN part
                     }
                     if main=0 then exit;
                     inc(ofs,pre);
                     count:=main;
                     buffer:=addr(ByteArray(buffer^)[pre]);
                   end;

                {
                  prepare position
                }
                realblock  := block^.xmsblock;
                realofs    := block^.offset;
                {
                  move the cursor by OFS/1024 Kilobytes
                }
                SeekXMS(realblock,realofs,ofs div 1024);
                {
                  add the rest of OFS to offset and convert it to Bytes
                }
                realoffset:=(longint(realofs)*1024) + (ofs mod 1024);
                {
                  now do the transfer (multiblock)
                }
                while count>0 do
                      begin
                        {
                          if the all bytes are in this block
                        }
                        if (realoffset+count)<=
                           (longint(realblock^.size)*1024) then
                           begin
                             with xmscore.xmsstruct do
                               begin
                                 length         := count;
                                 sourceh        := 0;
                                 source         := buffer;
                                 desth          := realblock^.handle;;
                                 longint(destination) := realoffset;
                               end;
                             xmscore.xmsmove;
                             count:=0;
                           end else
                           begin
                             with xmscore.xmsstruct do
                               begin
                                 length         := (longint(realblock^.size)
                                                    *1024)-realoffset;
                                 sourceh        := 0;
                                 source         := buffer;
                                 desth          := realblock^.handle;;
                                 longint(destination) := realoffset;
                                 dec(count,length);
                                 dec(realoffset,
                                     longint(realblock^.size)*1024-length);
                                 {
                                   move the pointer
                                 }
                                 buffer:=addr(ByteA(buffer^)[length]);
                               end;
                             xmscore.xmsmove;
                             {
                               select next block
                             }
                             realblock:=realblock^.next;
                           end;

                      end;

              end;

 Begin

   {
     we are under MS-Windows or 1 byte transfer capable himem.sys
   }
   xmscap:=4;

 end.











