{

  System 2001 v.6
  --------------------------------------------------------------------------
  (C) 1997-2001, Rozsnyo electronics and software


      [G-PATH technology] GRAPHIC DRIVERS UNIT


  Programmed by:      Rozsnyo Daniel

  Version:            1.20.001 C
  Release date:       2001-11-19

  Copyright:          (C) 2001, Rozsnyo electronics and software

}

Unit GDrivers;

Interface

  {no$DEFINE debug}

  Uses AdvMM,Types;

  {
    multiple driver handling
  }
  Function  GetDriversCount : integer;
  Function  GetDriversName(i:word)  : string;
  Procedure ChooseDriver(i:word);
  {
    register external drivers
  }
  type

    TDriverInfo   = record
                      setup : procedure;
                      name  : string[32];
                    end;

  Procedure RegisterDriver(inf:TDriverInfo);
  procedure SetupDriver(mc,mi,sm,pb,pl:pointer);

  {
    graphic driver link part
  }
  type
    TModeInfo     = TGP_ModeInfo;
                    {
                    record
                      width    : word;
                      height   : word;
                      bitcount : byte;
                    end;
                    }


    TGetModeCount = procedure(var ct:word);
    TGetModeInfo  = procedure(m:word;var inf:TModeInfo);
    TSetMode      = procedure(m:word);
    TProcessBlock = procedure(x,y,a,b,x0,y0,bpl:integer;mb:PAmmPart;f:byte);
    TProcessLine  = procedure(x,y,l:integer;p:pointer;f:byte);

  var
    GET_MODECOUNT { mc } : TGetModeCount;
    GET_MODEINFO  { mi } : TGetModeInfo;
    SET_MODE      { sm } : TSetMode;
    PROCESS_BLOCK { pb } : TProcessBlock;
    PROCESS_LINE  { pl } : TProcessLine;

  {
    universal graphic driver - process
  }
  procedure GD_HANDLE;

Implementation

  Uses SysCalls;

  Const

    MAX_DRIVERS  = 16;

    NO_DRIVER    = $FFFF;
    NO_NAME      = 'n/a';

  Var

    {
      currently used driver. if Used_driver = NO_DRIVER, then no driver
      is active
    }
    Used_driver  : word;

    {
      additional data
    }
    modecount    : word;
    modeinfo     : TModeInfo;

    NUM_DRIVERS  : word;
    DRIVERS      : array[0..MAX_DRIVERS-1] of TDriverInfo;

  procedure NullProc; assembler;
            asm
              {
                Procedure which does nothing
              }
            end;

  Function  GetDriversCount : integer;
            begin
              {
                Return the count of drivers
              }
              GetDriversCount := NUM_DRIVERS;
            end;

  Function  GetDriversName(i:word)  : string;
            begin
              {
                Return the name of driver (i=0..count-1)
              }
              if i<NUM_DRIVERS
                 then {
                        driver (i) exists - return the name from table
                      }
                      GetDriversName:=DRIVERS[i].name
                 else {
                        driver (i) doesn't exists
                      }
                      GetDriversName:=NO_NAME;
            end;

  procedure SetupDriver(mc,mi,sm,pb,pl:pointer);
            begin
              {
                load the parameters to global procedure pointers
                 => makes available the driver
              }
              @GET_MODECOUNT := mc;
              @GET_MODEINFO  := mi;
              @SET_MODE      := sm;
              @PROCESS_BLOCK := pb;
              @PROCESS_LINE  := pl;
            end;

  Procedure RegisterDriver(inf:TDriverInfo);
            begin
              {
                add a driver to drivers table
              }
              if NUM_DRIVERS < MAX_DRIVERS then
                 begin
                   {
                     insert driver to table of drivers
                   }
                   DRIVERS[num_drivers]:=inf;
                   inc(Num_drivers);
                   {$IFDEF debug}
                     Writeln('Graphic driver registred ['+inf.name+']');
                   {$ENDIF}
                 end else
                 begin
                   {
                     no space for driver's info
                   }
                 end;
            end;

  Procedure ChooseDriver(i:word);
            begin
              {
                Initialize the driver (i=0..count-1)
              }
              if i<NUM_DRIVERS then
                 begin
                   {
                     call the setup proedure of driver
                   }
                   DRIVERS[i].Setup;
                 end else
                 begin
                   {
                     driver (i) doesn't exists
                   }
                 end;
            end;

  procedure GD_HANDLE;
            var {
                   Local message
                }
                LM : TMessage;
            begin
              LM:=Message;
              case LM.mess of

                 PR_INIT :
                   begin
                     {
                       set global variable - faster link to graphic driver
                       and it is also checked by GUI Loader
                     }
                     ToGraph := Process.ID;
                   end;

                 PR_Close,
                 PR_Hibernate,
                 PR_Kill:
                   begin
                     {
                       standard response to Close/Kill signal
                     }
                     SendMessage( Process.ID, ToKernel,
                                  KE_StopProcess, 0, nil);
                   end;

                 GP_DRIVERSCOUNT:
                   begin
                     {
                       get the number of driver
                     }
                     SendMessage( Process.ID, LM.From, LM.Mess,
                                  2,addr(NUM_DRIVERS));
                   end;

                 GP_DRIVERSNAME:
                   begin
                     {
                       get the name of driver (W)
                     }
                     SendMessage( Process.ID, LM.From, LM.Mess,
                                  sizeof(DRIVERS[0].name),
                                  addr(DRIVERS[word(LM.data^)].NAME));
                   end;

                 GP_USEDRIVER:
                   begin
                     {
                       use driver (w) and return number of modes (w)
                     }
                     DRIVERS[word(LM.data^)].SETUP;
                     {
                       send back an event
                     }
                     SendMessage( Process.ID, LM.From, LM.Mess, 0, nil);
                   end;

                 GP_MODECOUNT:
                   begin
                     {
                       return number of modes (w)
                     }
                     GET_MODECOUNT(modecount);
                     SendMessage( Process.ID, LM.From, LM.Mess,
                                  2, addr(modecount));
                   end;

                 GP_MODEINFO:
                   begin
                     {
                       return information about mode (size.x,size.y,bpp)
                     }
                     GET_MODEINFO(word(LM.data^),modeinfo);
                     SendMessage( Process.ID, LM.From, LM.Mess,
                                  sizeof(modeinfo), addr(modeinfo));
                   end;

                 GP_SETMODE:
                   begin
                     {
                       return information about mode (size.x,size.y,bpp)
                     }
                     SET_MODE(word(LM.data^));
                     SendMessage( Process.ID, LM.From, LM.Mess, 0, nil);
                   end;

              end;
            end;

Begin
  {
    initialize the unit
  }
  Used_Driver    := NO_DRIVER;
  Num_Drivers    := 0;
  SetupDriver(addr(NullProc),
              addr(NullProc),
              addr(NullProc),
              addr(NullProc),
              addr(NullProc));
end.

