unit MemImage;

  { image manipulating in memory }

interface
  uses  advmm,types;

  const
       {3D shade}
        WinsBackColorInitial  = $C0C0C0;
        WinsDarkColorInitial  = $404040;
        WinsLightColorInitial = $f0f0f0;

       {Color Shift}
        WinsColorShift = $0000A0;
        WinsColorShiftRatio = $40;

       {Title}
        WinsTitleAL    = $008000; { active, left side of title bar }
        WinsTitleAR    = $000080; { active, right side of title bar }
        WinsTitleAT    = $40FF40; { active, text color }
        WinsTitleAI    = $4040FF; { active, icons, normal  }
        WinsTitleAO    = $C0C0FF; { active, icons, onmouse }
        WinsTitleAP    = $FF8080; { active, icons, pressed }

        WinsTitleDL    = $A0A0A0;
        WinsTitleDR    = $505050;
        WinsTitleDT    = $C0C0C0;
        WinsTitleDI    = $909090;

       {Help}
        WinsHelpColor  = $A0D0A0;

  var
        WinsBackColor  : Tcolor;
        WinsDarkColor  : Tcolor;
        WinsLightColor : Tcolor;


  type  TmemPalette = array[0..255] of TRGBColor;
        PmemPalette = ^TmemPalette;
        Tmemimage = object
                      width  : word;
                      height : word;
                      bits   : byte;
                      mblock : PammPart;
                      bpl    : longint;
                      bpp    : byte;

                      color  : TColor;
                      alpha  : byte;

                      palette: PmemPalette;

                      uMin   : Tpoint;
                      uMax   : Tpoint;

                      constructor init;

                      function    Alloc(x,y:word;b:byte):boolean;
                      procedure   LoadFromBmp(fn:string);

                      procedure   Fill(c:Tcolor);
                      procedure   FillMask(a:byte);
                      procedure   AutoSetAlpha(c:Tcolor);

                      procedure   PutPixel(x,y:integer);
                      function    GetPixel(x,y:integer):Tcolor;

                      procedure   PutPixelA(x,y:integer;a:byte);

                      procedure   SetColor(c:Tcolor);
                      procedure   SetColor32(c:Tcolor;a:byte);
                      function    GetColor:TColor;

                      procedure   MakeAfromRGB(destcol:TColor);
                                  { counts gray value from RGB parts
                                    and sets the mask to this value
                                    the color will be set to DestCol }

                      procedure   Line(x1,y1,x2,y2:integer);
                      procedure   Circle(x,y:integer;r:word);
                      procedure   FillCircle(x,y:integer;r:word);
                      procedure   Bar(x1,y1,x2,y2:integer);
                      procedure   Box(x,y,a,b:integer);
                      procedure   Rect(x1,y1,x2,y2:integer);
                      procedure   Frame(x,y,a,b:integer);

                      procedure   LinearShadeBoxX(x1,y1,x2,y2:integer;c1,c2:Tcolor);
                      procedure   Colorize(x1,y1,x2,y2:integer;c:Tcolor;a:byte);
                      procedure   ColorizeBOX(x1,y1,a,b:integer;c:Tcolor;al:byte);

                      procedure   DIPt  (x,y,a,b:integer;asp:byte);
                      procedure   UNDIPt(x,y,a,b:integer;asp:byte);
                      procedure   DIP   (x,y,a,b:integer;asp:byte);
                      procedure   UNDIP (x,y,a,b:integer;asp:byte);

                      Function    DrawChar(x,y:integer;n:char):byte;
                      Function    DrawString(x,y:integer;n:string):word;
                      Function    DrawStringCenter(x,y:integer;n:string):word;
                      Function    DrawStringRight(x,y:integer;n:string):word;
                      Procedure   DrawStringBlock(x,y,l:integer;n:string);

                      Function    DrawStringLen(x,y,l:integer;var n:string):word;
                      Function    DrawStringCenterLen(x,y,l:integer;var n:string):word;

                      procedure   PlaceBitmapCenter(fn:string);
                      procedure   PlaceBitmap(fn:string);
                      procedure   PlaceAlpha(fn:string);

                      procedure   PlaceBmp(x,y:integer;fn:string);
                      procedure   PlaceBmpT(x,y:integer;fn:string;c:Tcolor);

                      procedure   SaveToBmp(x,y:integer;a,b:word;fn:string);

                      procedure   PlaceStamp(x,y:integer;col:TColor;fn:string);

                      procedure   PlaceImage(dx,dy,sx,sy,a,b:integer;s:pointer);
                      procedure   PlaceImageDis(dx,dy,sx,sy,a,b:integer;s:pointer);

                      procedure   GetLine(x,y:integer;w:word;p:PRGBcolorArr);
                      procedure   PutLine(x,y:integer;w:word;p:PRGBcolorArr);

                      procedure   PutAlpha(x,y:integer;w:word;p:Pbuff);
                      procedure   GetAlpha(x,y:integer;w:word;p:Pbuff);

                      procedure   DrawLine (sx,sy,ix,iy:integer;l:word);
                      procedure   DrawBlock(sx,sy,ix,iy:integer;a,b:word);

                      procedure   Prepare;
                      function    NeedRedraw:boolean;

                      destructor  done;
                    end;

        Pmemimage = ^Tmemimage;

  function  Mix(c1,c2:Tcolor;aspect:byte):Tcolor;
  function  MakeColor(var r,g,b:byte):Tcolor;
  function  Ceq(a,b:TrgbColor):boolean;
  procedure ConvertColor(c:Tcolor;var rgb:TRGBcolor);
  function  Mid(c1,c2:Tcolor):Tcolor;

  function  HSLtoCOLOR(hue,sat,lig:byte):TColor;

  procedure ColorNormal;
  procedure ColorShift;

implementation
  uses bmp,font16,convert;

  var  WinsCol : record
                   light : Tcolor;
                   back  : Tcolor;
                   dark  : Tcolor;
                 end;

  procedure   ColorNormal;
              begin
                winsLightColor:=WinsLightColorInitial;
                winsBackColor :=WinsBackColorInitial;
                winsDarkColor :=WinsDarkColorInitial;
              end;

  procedure   ColorShift;
              begin
                winsLightColor:=mix(WinsLightColorInitial,
                                    WinsColorShift,
                                    WinsColorShiftRatio);
                winsBackColor :=mix(WinsBackColorInitial,
                                    WinsColorShift,
                                    WinsColorShiftRatio);
                winsDarkColor :=mix(WinsDarkColorInitial,
                                    WinsColorShift,
                                    WinsColorShiftRatio);
              end;

  function    MakeColor(var r,g,b:byte):Tcolor;
              begin
                MakeColor:=longint(r)*65536+longint(g)*256+b;
              end;

  function    Ceq(a,b:TrgbColor):boolean;
              begin
                ceq:=(a.r=b.r) and
                     (a.g=b.g) and
                     (a.b=b.b);
              end;

  procedure   ConvertColor(c:Tcolor;var rgb:TRGBcolor);
              begin
                rgb.r:=(c shr 16) and $FF;
                rgb.g:=(c shr  8) and $FF;
                rgb.b:=(c shr  0) and $FF;
              end;

   function  HSLtoCOLOR(hue,sat,lig:byte):TColor;
             var red,grn,blu: byte;
                 basic      : TColor;
             function huefn(h:word):byte;
                      var c : byte;
                      begin
                        c:=(h mod 256);
                        if c>(512 div 3) then
                           begin
                             huefn:=0;
                             exit;
                           end;
                        if c>(128) then
                           begin
                             huefn:=((512 div 3)-c)*6;
                             exit;
                           end;
                        if c>=(256 div 6) then
                           begin
                             huefn:=255;
                             exit;
                           end;
                        huefn:=c*6;
                      end;
             begin
               red:=huefn(hue+(256 div 3));
               grn:=huefn(hue);
               blu:=huefn(hue+(512 div 3));
               basic:=MakeColor(red,grn,blu);
               basic:=mix($808080,basic,sat);
               if lig<128 then basic:=mix($000000,basic,lig*2);
               if lig>128 then basic:=mix($FFFFFF,basic,(255-lig)*2);
               HSLtoCOLOR:=basic;
             end;
  function  Mid(c1,c2:Tcolor):Tcolor;
            var t : longint;     { longint  LLLL HHHH - bb gg rr 00 }
            begin
              asm
                       mov si,0
                @next: xor ax,ax
                       xor cx,cx
                       mov al,byte ptr [c1+si]
                       mov cl,byte ptr [c2+si]
                       add ax,cx
                       shr ax,1
                       mov byte ptr[t+si],al
                       inc si
                       cmp si,3
                       jne @next
              end;
              mid:=t;
            end;

  function  Mix(c1,c2:Tcolor;aspect:byte):Tcolor;
            var t : longint;     { longint  LLLL HHHH - bb gg rr 00 }
            begin
              asm
                       mov si,0
                       mov dl,aspect
                       mov dh,$FF
                       sub dh,dl
                @next: mov al,byte ptr [c1+si]
                       mul dh
                       mov cx,ax
                       mov al,byte ptr [c2+si]
                       mul dl
                       add ax,cx
                       mov byte ptr[t+si],ah
                       inc si
                       cmp si,3
                       jne @next
              end;
              mix:=t;
            end;

  Procedure MixRGBA(var s:TrgbColor;d:Trgbacolor);
            begin
              s.r:=s.r+(longint(d.r-s.r)*d.a) div 255;
              s.g:=s.g+(longint(d.g-s.g)*d.a) div 255;
              s.b:=s.b+(longint(d.b-s.b)*d.a) div 255;
            end;


{
                        LoadMask -> PlaceAlpha - universal

  procedure   Tmemimage.LoadMask(fn:string);
              type mb = array[0..159] of byte;
              var f:file;
                  b:mb;
                  l:longint;
                  x:integer;
              begin
                assign(f,fn);
                reset(f,1);
                for l:=0 to 199 do
                    begin
                      blockread(f,b,160);
                      putalpha(0,l,160,addr(b))
                    end;
                close(f);
              end;
}

  procedure   TMemImage.MakeAfromRGB(destcol:TColor);
              var x,y : integer;
                  buf : PRGBAcolorarr;
                  dc  : TRGBcolor;
              begin
                if bits<>32 then exit;
                if maxavail<bpl then exit;
                convertcolor(destcol,dc);
                getmem(buf,bpl);
                for y:=0 to height-1 do
                    begin
                      xms.read(mblock,y*bpl,bpl,pointer(buf));
                      for x:=0 to width-1 do
                          begin
                            buf^[x].a:=(integer(buf^[x].g)*59+
                                        integer(buf^[x].b)*11+
                                        integer(buf^[x].r)*30) div 100;
                            buf^[x].r:=dc.r;
                            buf^[x].g:=dc.g;
                            buf^[x].b:=dc.b;
                          end;
                      xms.write(mblock,y*bpl,bpl,pointer(buf));
                    end;
                freemem(buf,bpl);
              end;

  procedure   TMemImage.ColorizeBOX(x1,y1,a,b:integer;c:Tcolor;al:byte);
              begin
                colorize(x1,y1,x1+a-1,y1+b-1,c,al)
              end;

  procedure   TMemImage.SaveToBmp(x,y:integer;a,b:word;fn:string);
              type bmfh=bmpfileheader;
              var h  : bmfh;
                  f  : file;
                  pt : PRGBColorArr;
                  l  : integer;
                  lw : word;

              begin
                if a<1 then exit;
                if b<1 then exit;
                if x>=width then exit;
                if y>=width then exit;
                lw:=(((a+1)*3) div 4)*4;
                with h do
                  begin
                    bmptype[0]:='B';
                    bmptype[1]:='M';
                    filesize:=longint(b)*lw+sizeof(bmfh);
                    reserv1:=0;
                    reserv2:=0;
                    dataofs:=54;
                    size:=40;
                    width:=a;
                    height:=b;
                    planes:=1;
                    bitcount:=24;
                    compress:=0;
                    imgsize:=0;
                    xpix2met:=$EC4;
                    ypix2met:=$EC4;
                    colorusd:=0;
                    colorimp:=0;
                 end;

                if maxavail<lw then exit;
                getmem(pt,lw);

                assign(f,fn);
                {$I-}
                rewrite(f,1);
                {$I+}
                if ioresult<>0 then
                   begin
                     freemem(pt,lw);
                     exit;
                   end;

                blockwrite(f,h,54);

                for l:=b-1 downto 0 do
                    begin
                      getline(x,y+l,a,pt);
                      {$I-}
                      blockwrite(f,pt^,lw);
                      {$I+}
                      if ioresult<>0 then break;
                    end;

                freemem(pt,lw);
                close(f);

              end;

  procedure   Tmemimage.FillMask(a:byte);
              var buf : PRGBAcolorArr;
                  x,y : longint;
              begin
                if bits<>32 then exit;
                if maxavail<bpl then exit;
                getmem(buf,bpl);
                for y:=0 to self.height-1 do
                    begin
                      xms.read(mblock,y*bpl,bpl,pointer(buf));
                      for x:=0 to self.width-1 do buf^[x].a:=a;
                      xms.write(mblock,y*bpl,bpl,pointer(buf));
                    end;
                freemem(buf,bpl);
              end;

  procedure   Tmemimage.AutoSetAlpha(c:Tcolor);
              var buf : PRGBAcolorArr;
                  x,y : longint;
                  r   : Trgbcolor;
              begin
                if bits<>32 then exit;
                if maxavail<bpl then exit;
                convertcolor(c,r);
                getmem(buf,bpl);
                for y:=0 to self.height-1 do
                    begin
                      xms.read(mblock,y*bpl,bpl,pointer(buf));
                      for x:=0 to self.width-1 do
                          if (buf^[x].r=r.r) and
                             (buf^[x].g=r.g) and
                             (buf^[x].b=r.b) then buf^[x].a:=0
                                             else buf^[x].a:=255;
                      xms.write(mblock,y*bpl,bpl,pointer(buf));
                    end;
                freemem(buf,bpl);
              end;


  constructor Tmemimage.init;
              begin
                width:=0;
                height:=0;
                bits:=0;
                mblock:=nil;
                alpha:=$80;
              end;

  function    Tmemimage.Alloc(x,y:word;b:byte):boolean;
              var memneeded : longint;
              begin
                alloc:=false;
                if mblock<>nil then xms.DeAlloc(mblock);
                if (b<>1) and
                   (b<>2) and
                   (b<>4) and
                   (b<>8) and
                   (b<>15) and
                   (b<>16) and
                   (b<>24) and
                   (b<>32) then
                   begin
                     bsod('Memory image allocation',
                          'size.x   = '+integertostring(x),
                          'size.y   = '+integertostring(y),
                          'bitcount = '+integertostring(b),
                          '');
                   end;
                bits:=b;
                case b of
                   1: if (x mod 8)=0
                         then bpl:=x div 8
                         else bpl:=(x div 8)+1;
                   4: if (x mod 2)=0
                         then bpl:=x div 2
                         else bpl:=(x div 2)+1;
                   8: bpl:=x;
                  15: bpl:=x*2;
                  16: bpl:=x*2;
                  24: bpl:=x*3;
                  32: bpl:=x*4;
                end;
                case b of
                   1: bpp:=0;
                   4: bpp:=0;
                   8: bpp:=1;
                  15: bpp:=2;
                  16: bpp:=2;
                  24: bpp:=3;
                  32: bpp:=4;
                end;
                memneeded:=bpl*y;
                if (memneeded mod 1024)>0 then inc(memneeded,1024);
                mblock:=xms.Alloc(memneeded div 1024);
                if mblock=nil then
                   begin
                     {not enough mmanager memory}
                     init;
                     exit;
                   end;
                alloc:=true;
                width:=x;
                height:=y;
              end;

  procedure   Tmemimage.LoadFromBmp(fn:string);
              begin
              end;

  procedure   Tmemimage.PlaceBitmapCenter(fn:string);
              var bmp : bmpfile;
                  xo  : integer;
                  yo  : integer;
                  b   : PRGBcolorArr;
                  i   : integer;
              begin
                bmp.init;
                bmp.openfile(fn);
                if not bmp.opened then
                   begin
                     {bmp file not found}
                     exit;
                   end;
                xo:=(self.width-bmp.size.x) div 2;
                yo:=(self.height-bmp.size.y) div 2;
                if (xo>=0) and (yo>=0) then
                   case bmp.bitperpixel of
                     24 : if maxavail>bmp.linesize then
                          begin
                            getmem(b,bmp.linesize);
                            for i:=bmp.size.y-1 downto 0 do
                                begin
                                  bmp.readline(i,b);
                                  self.putline(xo,yo+i,bmp.size.x,b);
                                end;
                            freemem(b,bmp.linesize);
                          end;
                   end;
                bmp.done;
              end;

  procedure   Tmemimage.PlaceBmp(x,y:integer;fn:string);
              var bmp : bmpfile;
                  b   : PRGBcolorArr;
                  i   : integer;
              begin
                bmp.init;
                bmp.openfile(fn);
                if not bmp.opened then
                   begin
                     {bmp file not found}
                     exit;
                   end;
                if (x<width) and (y<height) then
                   if maxavail>bmp.linesize then
                      begin
                        getmem(b,3*bmp.size.x);
                        for i:=bmp.size.y-1 downto 0 do
                            begin
                              bmp.readline24(i,b);
                              self.putline(x,y+i,bmp.size.x,b);
                            end;
                        freemem(b,3*bmp.size.x);
                      end;
                bmp.done;
              end;

  procedure   Tmemimage.PlaceBmpT(x,y:integer;fn:string;c:Tcolor);
              var bmp : bmpfile;
                  b   : PRGBcolorArr;
                  i   : integer;
                  j   : integer;
              begin
                bmp.init;
                bmp.openfile(fn);
                if not bmp.opened then
                   begin
                     {bmp file not found}
                     exit;
                   end;
                if (x<width) and (y<height) then
                   case bmp.bitperpixel of
                     24 : if maxavail>bmp.linesize then
                          begin
                            getmem(b,bmp.linesize);
                            for i:=bmp.size.y-1 downto 0 do
                                begin
                                  bmp.readline(i,b);
                                  for j:=0 to bmp.size.x-1 do
                                      begin
                                        self.color:=makecolor
                                          ( b^[j].r,b^[j].g,b^[j].b);
                                        if self.color<>c then
                                           self.putpixel(x+j,y+i);
                                      end;
                                end;
                            freemem(b,bmp.linesize);
                          end;
                   end;
                bmp.done;
              end;

  procedure   Tmemimage.PlaceBitmap(fn:string);
              var bmp : bmpfile;
                  b   : PRGBcolorArr;
                  i   : integer;
              begin
                bmp.init;
                bmp.openfile(fn);
                if not bmp.opened then
                   begin
                     {bmp file not found}
                     exit;
                   end;
                 case bmp.bitperpixel of
                   24 : if maxavail>bmp.linesize then
                        begin
                          getmem(b,bmp.linesize);
                          for i:=bmp.size.y-1 downto 0 do
                              begin
                                bmp.readline(i,b);
                                self.putline(0,i,bmp.size.x,b);
                              end;
                          freemem(b,bmp.linesize);
                        end;
                 end;
                bmp.done;
              end;

  procedure   TMemImage.PlaceStamp(x,y:integer;col:TColor;fn:string);
              var bmp : bmpfile;
                  b   : PBuff;
                  p   : PRGBcolorArr;
                  i   : integer;
                  j   : integer;
                  c   : TrgbColor;
              begin
                ConvertColor(col,c);
                bmp.init;
                bmp.openfile(fn);
                if not bmp.opened then
                   begin
                     {bmp file not found}
                     bmp.done;
                     exit;
                   end;
                 case bmp.bitperpixel of
                   8 : if maxavail>bmp.linesize then
                        begin
                          getmem(b,bmp.linesize);
                          getmem(p,bmp.linesize*3);
                          for i:=bmp.size.y-1 downto 0 do
                              begin
                                bmp.readline(i,pointer(b));
                                self.getline(x,y+i,bmp.size.x,p);
                                for j:=0 to bmp.size.x-1 do
                                    begin
                                      p^[j].r:=c.r+longint(p^[j].r-c.r)*b^[j] div 255;
                                      p^[j].g:=c.g+longint(p^[j].g-c.g)*b^[j] div 255;
                                      p^[j].b:=c.b+longint(p^[j].b-c.b)*b^[j] div 255;
                                    end;
                                self.putline(x,y+i,bmp.size.x,p);
                              end;
                          freemem(p,bmp.linesize*3);
                          freemem(b,bmp.linesize);
                        end;
                 end;
                bmp.done;
              end;

  procedure   Tmemimage.PlaceAlpha(fn:string);
              var bmp : bmpfile;
                  b   : Pbuff;
                  i   : integer;
              begin
                if bits<32 then exit;
                bmp.init;
                bmp.openfile(fn);
                if not bmp.opened then
                   begin
                     {bmp file not found}
                     exit;
                   end;
                 case bmp.bitperpixel of
                   8 : if maxavail>bmp.linesize then
                        begin
                          getmem(b,bmp.linesize);
                          for i:=bmp.size.y-1 downto 0 do
                              begin
                                bmp.readline(i,pointer(b));
                                self.putalpha(0,i,bmp.size.x,b);
                              end;
                          freemem(b,bmp.linesize);
                        end;
                 end;
                bmp.done;
              end;

  procedure   Tmemimage.Fill(c:Tcolor);
              var b : PRGBcolorArr;
                  d : PRGBAcolorArr;
                  i : integer;
                  r : TrgbColor;
                  a : Trgbacolor;
              begin
                case bits of
                  8  : ;
                  24 : begin
                         if maxavail<bpl then
                            begin
                              {no mem for buffer}
                              exit;
                            end;
                         getmem(b,bpl);
                         convertcolor(c,r);
                         for i:=0 to self.width-1 do
                             b^[i]:=r;
                         for i:=0 to self.height-1 do
                             PutLine(0,i,self.width,b);
                         freemem(b,bpl);
                       end;
                  32 : begin
                         if maxavail<bpl then
                            begin
                              {no mem for buffer}
                              exit;
                            end;
                         getmem(d,bpl);
                         convertcolor(c,r);
                         a.r:=r.r;
                         a.g:=r.g;
                         a.b:=r.b;
                         a.a:=alpha;
                         for i:=0 to self.width-1 do
                             d^[i]:=a;
                         for i:=0 to self.height-1 do
                             xms.write(mblock,bpl*i,bpl,pointer(d));
                         freemem(d,bpl);
                       end;
                end;
              end;

  procedure   Tmemimage.GetLine(x,y:integer;w:word;p:PRGBcolorArr);
              var b:Pbuff;
                  a:PRGBAcolorArr;
                  i:integer;
                  c1:TRGBcolor;
                  c2:TRGBAcolor;
              begin
                if x+w>width then w:=width-x;
                if x<0 then
                   begin
                     w:=w-(0-x);
                     p:=addr(p^[-x]);
                     x:=0;
                   end;
                if x>=width  then exit;
                if y>=height then exit;
                case bits of
                   1 : begin
                         if maxavail<bpl then exit;
                         getmem(b,bpl);
                         xms.read(mblock,y*bpl,bpl,b);
                         for i:=x to x+w-1 do
                             p^[i-x]:=palette^[(b^[i div 8] shr (i mod 8)) and 1];
                         freemem(b,bpl);
                       end;
                   4 : begin
                         if maxavail<bpl then exit;
                         getmem(b,bpl);
                         xms.read(mblock,y*bpl,bpl,b);
                         for i:=x to x+w-1 do
                             if (x mod 2)=0
                                then p^[i-x]:=palette^[b^[i div 2] shr 4]
                                else p^[i-x]:=palette^[b^[i div 2] and 15];
                         freemem(b,bpl);
                       end;
                   8 : begin
                         if maxavail<w then exit;
                         getmem(b,w);
                         xms.read(mblock,y*bpl+x,w,b);
                         for i:=0 to w-1 do
                             p^[i]:=palette^[b^[i]];
                         freemem(b,w);
                       end;
                  24 : xms.read(mblock,bpl*y+x*bpp,w*bpp,pointer(p));
                  32 : begin
                         getmem(a,w*bpp);
                         xms.read(mblock,bpl*y+x*bpp,w*bpp,pointer(a));

                               {
                               p^[i].b:=p^[i].b+(longint(a^[i].b-p^[i].b)*a^[i].a) div 255;
                               }

                         asm
                           push ds
                           mov cx,w
                           { number of pixels = w  }
                           mov di,word ptr a
                           mov ax,word ptr a+2
                           mov es,ax
                           { a= ES:SI }
                           mov si,word ptr p
                           mov ax,word ptr p+2
                           mov ds,ax
                           { p= DS:SI }
                           @NEW:
                                 { count the color, p=3, a=4 }
                                 { p0=p0+((a0-p0)*a3/255)    }

                                 xor ax,ax
                                 mov al,byte ptr es:[di]
                                 {ax:=a0}
                                 xor bx,bx
                                 mov bl,byte ptr ds:[si]
                                 {bx:=p0}
                                 sub ax,bx
                                 {ax:=a0-p0}
                                 xor bx,bx
                                 mov bl,byte ptr es:[di+3]
                                 {bx:=alpha}
                                 xor dx,dx
                                 imul bx
                                 {dx:ax:=(a0-p0)*a3}
                                 mov bx,255
                                 idiv bx
                                 {ax:=(a0-p0)*a3/255}
                                 xor bx,bx
                                 mov bl,byte ptr ds:[si]
                                 {bx:=p0}
                                 add bx,ax
                                 mov byte ptr ds:[si],bl

                                 xor ax,ax
                                 mov al,byte ptr es:[di+1]
                                 {ax:=a0}
                                 xor bx,bx
                                 mov bl,byte ptr ds:[si+1]
                                 {bx:=p0}
                                 sub ax,bx
                                 {ax:=a0-p0}
                                 xor bx,bx
                                 mov bl,byte ptr es:[di+3]
                                 {bx:=alpha}
                                 xor dx,dx
                                 imul bx
                                 {dx:ax:=(a0-p0)*a3}
                                 mov bx,255
                                 idiv bx
                                 {ax:=(a0-p0)*a3/255}
                                 xor bx,bx
                                 mov bl,byte ptr ds:[si+1]
                                 {bx:=p0}
                                 add bx,ax
                                 mov byte ptr ds:[si+1],bl

                                 xor ax,ax
                                 mov al,byte ptr es:[di+2]
                                 {ax:=a0}
                                 xor bx,bx
                                 mov bl,byte ptr ds:[si+2]
                                 {bx:=p0}
                                 sub ax,bx
                                 {ax:=a0-p0}
                                 xor bx,bx
                                 mov bl,byte ptr es:[di+3]
                                 {bx:=alpha}
                                 xor dx,dx
                                 imul bx
                                 {dx:ax:=(a0-p0)*a3}
                                 mov bx,255
                                 idiv bx
                                 {ax:=(a0-p0)*a3/255}
                                 xor bx,bx
                                 mov bl,byte ptr ds:[si+2]
                                 {bx:=p0}
                                 add bx,ax
                                 mov byte ptr ds:[si+2],bl

                                 { increment pointers }
                                 add si,3
                                 cmp si,16
                                 jb  @NX1
                                 sub si,16
                                 mov ax,ds
                                 inc ax
                                 mov ds,ax
                           @NX1:
                                 add di,4
                                 cmp di,16
                                 jb  @NX2
                                 sub di,16
                                 mov ax,es
                                 inc ax
                                 mov es,ax
                           @NX2:
                                 dec cx
                                 jnz @new
                           pop ds
                         end;

                         freemem(a,w*bpp);
                       end;
                end;
              end;

  procedure   Tmemimage.PutLine(x,y:integer;w:word;p:PRGBcolorArr);
              var b:PRGBAcolorArr;
                  i:integer;
              begin
                if x+w>width then w:=width-x;
                if x>=width  then exit;
                if y>=height then exit;
                case bits of
                   8 : xms.write(mblock,bpl*y+x,w,pointer(p));
                  24 : xms.write(mblock,bpl*y+x*bpp,w*bpp,pointer(p));
                  32 : begin
                         getmem(b,w*bpp);
                         xms.read(mblock,bpl*y+x*bpp,w*bpp,pointer(b));
                         for i:=0 to w-1 do
                             begin
                               b^[i].r:=p^[i].r;
                               b^[i].g:=p^[i].g;
                               b^[i].b:=p^[i].b;
                             end;
                         xms.write(mblock,bpl*y+x*bpp,w*bpp,pointer(b));
                         freemem(b,w*bpp);
                       end;
                end;
                if x<umin.x then umin.x:=x;
                if y<umin.y then umin.y:=y;
                if (x+w-1)>umax.x then umax.x:=x+w-1;
                if y>umax.y then umax.y:=y;
              end;

  procedure   TMemImage.PutPixelA(x,y:integer;a:byte);
              var b:TRGBAcolor;
              begin
                if bits<32 then exit;
                if y>=height then exit;
                if x>=width then exit;
                if x<0 then exit;
                if y<0 then exit;
                xms.read(mblock,bpl*y+x*4,4,addr(b));
                b.a:=a;
                xms.write(mblock,bpl*y+x*4,4,addr(b));
              end;

  procedure   TMemImage.SetColor32(c:Tcolor;a:byte);
              var b: TRGBcolor;
              begin
                alpha:=a;
                setcolor(c);
              end;

  procedure   Tmemimage.PutAlpha(x,y:integer;w:word;p:Pbuff);
              var b : PRGBAcolorArr;
                  i : integer;
              begin
                if bits<32 then exit;
                if w<1 then exit;
                if y>=height then exit;
                if x>=width then exit;
                if x<0 then exit;
                if y<0 then exit;
                if (x+w)>width then w:=width-x;
                if maxavail<w*4 then
                   begin
                     {no memory}
                     exit;
                   end;
                getmem(b,w*4);
                xms.read(mblock,bpl*y+x*4,w*4,pointer(b));
                for i:=0 to w-1 do
                    b^[i].a:=p^[i];
                xms.write(mblock,bpl*y+x*4,w*4,pointer(b));
                freemem(b,w*4);
              end;

  procedure   Tmemimage.GetAlpha(x,y:integer;w:word;p:Pbuff);
              var b : PRGBAcolorArr;
                  i : integer;
              begin
                if bits<32 then exit;
                if w<1 then exit;
                if y>=height then exit;
                if x>=width then exit;
                if x<0 then exit;
                if y<0 then exit;
                if (x+w)>width then w:=width-x;
                if maxavail<w*4 then
                   begin
                     {no memory}
                     exit;
                   end;
                getmem(b,w*4);
                xms.read(mblock,bpl*y+x*4,w*4,pointer(b));
                for i:=0 to w-1 do
                    p^[i]:=b^[i].a;
                freemem(b,w*4);
              end;

  procedure   Tmemimage.DrawLine(sx,sy,ix,iy:integer;l:word);
              var b: PRGBcolorArr;
              begin
                if l<1 then exit;
                if maxavail<(bpp*l) then
                   begin
                     {no mem buffer};
                     exit;
                   end;
                getmem(b,bpp*l);
                self.getline(ix,iy,l,b);
                {gr24.putline(sx,sy,l,b);}
                freemem(b,bpp*l);
              end;

  procedure   Tmemimage.DrawBlock(sx,sy,ix,iy:integer;a,b:word);
              var u : PRGBcolorArr;
                  l : integer;
                  h : boolean;
              begin
                if b<1 then exit;
                if a<1 then exit;
                {if sx>max.x then exit;
                if sy>max.y then exit;
                if (sx+a-1)>max.x then
                   a:=max.x-sx+1;
                if (sy+b-1)>max.y then
                   b:=max.y-sy+1;
                if sx<min.x then
                   begin
                     a:=a-(min.x-sx);
                     ix:=ix+(min.x-sx);
                     sx:=min.x;
                   end;
                if sy<min.y then
                   begin
                     b:=b-(gr24.min.y-sy);
                     iy:=iy+(gr24.min.y-sy);
                     sy:=gr24.min.y;
                   end;
                }
                if maxavail<(bpp*a) then
                   begin
                     {no mem buffer};
                     exit;
                   end;
                getmem(u,bpp*a);
                for l:=0 to b-1 do
                    begin
                      self.getline(ix,iy+l,a,u);
                      {gr24.putline(sx,sy+l,a,u);}
                    end;
                freemem(u,bpp*a);
              end;

  procedure   Tmemimage.PutPixel(x,y:integer);
              begin
                if x<0 then exit;
                if y<0 then exit;
                if x>=width then exit;
                if y>=height then exit;
                case bits of
                  8,24 : xms.write(mblock,bpl*y+x*bpp,bpp,addr(self.color));
                  32   : xms.write(mblock,bpl*y+x*bpp,4,addr(self.color));
                end;
                if x<umin.x then umin.x:=x;
                if y<umin.y then umin.y:=y;
                if x>umax.x then umax.x:=x;
                if y>umax.y then umax.y:=y;
              end;

  function    Tmemimage.GetPixel(x,y:integer):Tcolor;
              var color24 : Tcolor;
                  color8  : byte;
              begin
                case bits of
                  8 : begin
                         xms.read(mblock,bpl*y+x*bpp,1,addr(color8));
                         getpixel:=color8;
                       end;
                  24,
                  32 : begin
                         xms.read(mblock,bpl*y+x*bpp,3,addr(color24));
                         getpixel:=color24 and $FFFFFF;
                       end;
                end;
              end;

  function    Tmemimage.GetColor:Tcolor;
              begin
                GetColor:=color and $00FFFFFF;
              end;

  procedure   Tmemimage.SetColor(c:Tcolor);
              begin
                color:=(longint(alpha) shl 24) or (c and $00FFFFFF);
              end;

  procedure   Tmemimage.Line(x1,y1,x2,y2:integer);
              var max : longint;
                  t   : longint;
                  b   : PRGBcolorArr;
                  ba  : PColorArr;
                  r   : TrgbColor;
              begin
                max:=abs(x1-x2);
                if y1=y2 then
                   begin
                     {
                       in one line
                     }
                     if x1=x2 then
                        begin
                          {
                            in one point
                          }
                          putpixel(x1,y1);
                          exit;
                        end;
                     inc(max);
                     if (bits=24) and
                        (maxavail>(max*bpp)) then
                        begin
                          {
                            if enought memory and 24 bit color
                          }
                          if x1>x2 then x1:=x2;
                          getmem(b,max*bpp);
                          ConvertColor(color,r);
                          for t:=0 to max-1 do
                              b^[t]:=r;
                          self.putline(x1,y1,max,b);
                          freemem(b,max*bpp);
                          if x1<umin.x then umin.x:=x1;
                          if (x1+max-1)>umax.x then umax.x:=x1+max-1;
                          if y1<umin.y then umin.y:=y1;
                          if y1>umax.y then umax.y:=y1;
                          exit;
                        end;
                     if (bits=32) and
                        (maxavail>(max*bpp)) then
                        begin
                          {
                            if enought memory and 32 bit color
                          }
                          if x1>x2 then x1:=x2;
                          getmem(ba,max*bpp);
                          for t:=0 to max-1 do
                              ba^[t]:=self.color;
                          xms.write(mblock,bpl*y1+bpp*x1,bpp*max,pointer(ba));
                          freemem(ba,max*bpp);
                          {
                          for t:=0 to max-1 do
                              putpixel(y1,x1+t);
                          }
                          if x1<umin.x then umin.x:=x1;
                          if (x1+max-1)>umax.x then umax.x:=x1+max-1;
                          if y1<umin.y then umin.y:=y1;
                          if y1>umax.y then umax.y:=y1;
                          exit;
                        end;
                   end;
                if max<abs(y1-y2) then max:=abs(y1-y2);
                for t:=0 to max do
                    self.putpixel(x1+round((t*(x2-x1))/max),
                                  y1+round((t*(y2-y1))/max));
              end;

  procedure   Tmemimage.Circle(x,y:integer;r:word);
              var d,xx,yy : integer;
              begin
                d:=3-(2*r);
                xx:=0;
                yy:=r;
                repeat
                  PutPixel(X+Xx,Y+Yy);
                  PutPixel(X+Xx,Y-Yy);
                  PutPixel(X-Xx,Y+Yy);
                  PutPixel(X-Xx,Y-Yy);
                  PutPixel(X+Yy,Y+Xx);
                  PutPixel(X+Yy,Y-Xx);
                  PutPixel(X-Yy,Y+Xx);
                  PutPixel(X-Yy,Y-Xx);
                  inc(xx);
                  if d<0 then d:=d+(4*xx)+6
                         else begin
                                d:=d+4*(xx-yy)+10;
                                yy:=yy-1;
                              end;
                until xx>yy;
              end;

  procedure   Tmemimage.FillCircle(x,y:integer;r:word);
              var d,xx,yy : integer;
              begin
                d:=3-(2*r);
                xx:=0;
                yy:=r;
                repeat
                  if xx<>0 then Line(x-xx,y+yy,x+xx,y+yy);
                  if xx<>0 then Line(x-xx,y-yy,x+xx,y-yy);
                  Line(x-yy,y+xx,x+yy,y+xx);
                  Line(x-yy,y-xx,x+yy,y-xx);
                  inc(xx);
                  if d<0 then d:=d+(4*xx)+6
                         else begin
                                d:=d+4*(xx-yy)+10;
                                yy:=yy-1;
                              end;
                until xx>yy;
              end;

  procedure   Tmemimage.Bar(x1,y1,x2,y2:integer);
              var i:integer;
              begin
                if x1>x2 then
                   begin
                     i:=x1;
                     x1:=x2;
                     x2:=i;
                   end;
                if y1>y2
                   then for i:=y2 to y1 do line(x1,i,x2,i)
                   else for i:=y1 to y2 do line(x1,i,x2,i);
              end;

  procedure   TmemImage.Colorize(x1,y1,x2,y2:integer;c:Tcolor;a:byte);
              var l:integer;
                  i:integer;
                  p:PRGBColorArr;
              procedure ColLine(x,y,l:integer;c:Tcolor;a:byte);
                        var i:integer;
                        begin
                          GetLine(x,y,l,p);
                          for i:=0 to l-1 do
                              ConvertColor(
                                 mix(MakeColor(p^[i].r,p^[i].g,p^[i].b),c,a),
                                 p^[i]);
                          PutLine(x,y,l,p);
                        end;
              begin
                if bits<24 then exit;
                l:=abs(x2-x1)+1;
                if x1>x2 then
                   begin
                     i:=x1;
                     x1:=x2;
                     x2:=i;
                   end;
                getmem(p,l*3);
                if y1>y2
                   then for i:=y2 to y1 do Colline(x1,i,l,c,a)
                   else for i:=y1 to y2 do Colline(x1,i,l,c,a);
                freemem(p,l*3);
              end;

  procedure   Tmemimage.LinearShadeBoxX(x1,y1,x2,y2:integer;c1,c2:Tcolor);
              var i:integer;
              begin
                if x2=0 then exit;
                for i:=x1 to (x1+x2-1) do
                    begin
                      setcolor(mix(c1,c2,round((i-x1)/x2*255)));
                      line(i,y1,i,y1+y2-1);
                    end;
              end;

  procedure   Tmemimage.Box(x,y,a,b:integer);
              begin
                self.bar(x,y,x+a-1,y+b-1);
              end;

  procedure   Tmemimage.Rect(x1,y1,x2,y2:integer);
              begin
                line(x1,y1,x2,y1);
                line(x1,y2,x2,y2);
                line(x1,y1,x1,y2);
                line(x2,y1,x2,y2);
              end;

  procedure   Tmemimage.Frame(x,y,a,b:integer);
              begin
                self.rect(x,y,x+a-1,y+b-1);
              end;

  destructor  Tmemimage.done;
              begin
                if mblock<>nil then xms.DeAlloc(mblock);
              end;

  procedure   TmemImage.DIPt(x,y,a,b:integer;asp:byte);
              begin
                setcolor(mix(winsbackcolor,winsdarkcolor,asp));
                line(x,    y,    x+a-2,y);
                line(x,    y+1,  x,    y+b-2);
                setcolor(mix(winsbackcolor,winslightcolor,asp));
                line(x,    y+b-1,x+a-2,y+b-1);
                line(x+a-1,y,    x+a-1,y+b-1);
              end;

  procedure   TmemImage.UNDIPt(x,y,a,b:integer;asp:byte);
              begin
                setcolor(mix(winsbackcolor,winslightcolor,asp));
                line(x,    y,    x+a-1,y);
                line(x,    y+1,  x,    y+b-1);
                setcolor(mix(winsbackcolor,winsdarkcolor,asp));
                line(x+1,  y+b-1,x+a-2,y+b-1);
                line(x+a-1,y+1,  x+a-1,y+b-1);
              end;

  procedure   TmemImage.DIP(x,y,a,b:integer;asp:byte);
              begin
                setcolor(mix(winsbackcolor,winsdarkcolor,asp));
                line(x,    y,    x+a-2,y);
                line(x,    y+1,  x,    y+b-2);
                setcolor(mix(winsbackcolor,winslightcolor,asp));
                line(x,    y+b-1,x+a-2,y+b-1);
                line(x+a-1,y,    x+a-1,y+b-1);
                setcolor(winsbackcolor);
                bar (x+1,  y+1,  x+a-2,y+b-2);
              end;

  procedure   TmemImage.UNDIP(x,y,a,b:integer;asp:byte);
              begin
                setcolor(mix(winsbackcolor,winslightcolor,asp));
                line(x,    y,    x+a-1,y);
                line(x,    y+1,  x,    y+b-1);
                setcolor(mix(winsbackcolor,winsdarkcolor,asp));
                line(x+1,  y+b-1,x+a-2,y+b-1);
                line(x+a-1,y+1,  x+a-1,y+b-1);
                setcolor(winsbackcolor);
                bar (x+1,  y+1,  x+a-2,y+b-2);
              end;

  procedure   Tmemimage.Prepare;
              begin
                umin.x:=width;
                umin.y:=height;
                umax.x:=-1;
                umax.y:=-1;
              end;

  function    Tmemimage.NeedRedraw:boolean;
              begin
                needredraw:=((umax.x-umin.x)>=0)and((umax.y-umin.y)>=0);
              end;

  Function    Tmemimage.DrawStringCenter(x,y:integer;n:string):word;
              var l:word;
              begin
                l:=lenstring(n);
                drawstring(x-(l div 2),y,n);
              end;

  Function    Tmemimage.DrawStringRight(x,y:integer;n:string):word;
              var l:word;
              begin
                l:=lenstring(n);
                drawstring(x-l,y,n);
              end;

  Function    Tmemimage.DrawChar(x,y:integer;n:char):byte;
              var c  : byte;
                  ms : word;
                  xo : word;
                  yo : word;
                  fd : array[0..15] of word;
              begin
                asm
                      mov ah,n
                      mov c,ah
                      mov ms,0
                end;
                if x>(width-1)  then exit;
                if y>(height-1) then exit;
                setcolor(fonts.color);
                for yo:=0 to 15 do fd[yo]:=fonts.data[fonts.used]^[c*16+yo];
                case fonts.zoom of
                  false : for yo:=0 to 15 do
                          for xo:=0 to 15 do
                              if ((fd[yo])and($8000 shr xo))>0
                                 then begin
                                        if (xo*fonts.xs)>ms then ms:=xo*fonts.xs;
                                        putpixel(x+xo,y+yo);
                                      end;
                  true  : for yo:=0 to 15 do
                          for xo:=0 to 15 do
                              if ((fd[yo])and($8000 shr xo))>0
                                 then begin
                                        if (xo*fonts.xs)>ms then ms:=xo*fonts.xs;
                                        box( x+xo*fonts.xs,
                                             y+yo*fonts.ys,
                                             fonts.xz,
                                             fonts.yz);
                                      end;
                end;
                drawchar:=ms;
              end;

  Function    Tmemimage.DrawString(x,y:integer;n:string):word;
              var i : byte;
                  l : word;
              begin
                if n[0]=#0 then
                   begin
                     drawstring:=0;
                     exit;
                   end;
                l:=0;
                for i:=1 to ord(n[0]) do l:=l+drawchar(x+l,y,n[i])+fonts.spacing;
                DrawString:=l-fonts.spacing;
              end;

  Function    Tmemimage.DrawStringLen(x,y,l:integer;var n:string):word;
              var s:string;
              begin
                s:=n;
                if l<8 then exit;
                if lenstring(s)>l then s:=s+'...';
                while lenstring(s)>l do s:=copy(s,1,length(s)-4)+'...';
                self.drawstring(x,y,s);
                n:=s;
              end;

  Function    Tmemimage.DrawStringCenterLen(x,y,l:integer;var n:string):word;
              var s:string;
              begin
                s:=n;
                if l<8 then exit;
                if lenstring(s)>l then s:=s+'...';
                while lenstring(s)>l do s:=copy(s,1,length(s)-4)+'...';
                self.drawstringcenter(x,y,s);
                n:=s;
              end;

  Procedure   TmemImage.DrawStringBlock(x,y,l:integer;n:string);
              var slova : array[0..63] of string[32];
                  medz  : integer;
                  i     : byte;
                  m     : byte;
                  dslov : integer;
                  xx    : integer;
                  xxx   : integer;
              begin
                if l<3 then exit;
                if n='' then exit;
                if pos(' ',n)=0 then
                   begin
                     drawstringcenter(x,y+l div 2,n);
                     exit;
                   end;
                m:=length(n);
                medz:=0;
                slova[0]:='';
                for i:=1 to m do
                    if n[i]=' ' then begin
                                       inc(medz);
                                       slova[medz]:='';
                                     end
                                else slova[medz]:=slova[medz]+n[i];
                dslov:=0;
                for i:=0 to medz do dslov:=dslov+lenstring(slova[i]);
                xx:=0;
                for i:=0 to medz do
                    begin
                      xxx:=round(i*(l-dslov)/medz);
                      drawstring(x+xx+xxx,y,slova[i]);
                      xx:=xx+lenstring(slova[i]);
                    end;
              end;

  procedure   TmemImage.PlaceImage(dx,dy,sx,sy,a,b:integer;s:pointer);
              var src : PmemImage;
                  y   : integer;
                  buf : PRGBcolorArr;
              begin
                src:=s;
                if mblock=nil then exit;
                if maxavail<a*3 then
                   begin
                     {No Mem Error}
                     exit;
                   end;
                getmem(buf,a*3);
                case src^.bits of
                  24 : for y:=0 to b-1 do
                           begin
                             src^.getline(sx,sy+y,a,buf);
                             self.Putline(dx,dy+y,a,buf);
                           end;
                  32 : for y:=0 to b-1 do
                           begin
                             self.getline(dx,dy+y,a,buf);
                             src^.getline(sx,sy+y,a,buf);
                             self.Putline(dx,dy+y,a,buf);
                           end;
                end;
                freemem(buf,a*3)
              end;

  procedure   TmemImage.PlaceImageDis(dx,dy,sx,sy,a,b:integer;s:pointer);
              var src : PmemImage;
                  y   : integer;
                  buf : PRGBcolorArr;
              procedure changebuff(b:PRGBcolorArr;l:integer);
                        var i:integer;
                            n:integer;
                        begin
                          for i:=0 to l-1 do
                              begin
                                n:=(integer(b^[i].g)*59+
                                    integer(b^[i].b)*11+
                                    integer(b^[i].r)*30) div 100;
                                b^[i].r:=(n+2*( winsbackcolor shr 16)) div 3;
                                b^[i].g:=(n+2*((winsbackcolor shr 8)and 255)) div 3;
                                b^[i].b:=(n+2*( winsbackcolor and 255)) div 3;
                              end;
                        end;
              begin
                src:=s;
                if mblock=nil then exit;
                if maxavail<a*3 then
                   begin
                     {No Mem Error}
                     exit;
                   end;
                getmem(buf,a*3);
                case src^.bits of
                  24 : for y:=0 to b-1 do
                           begin
                             src^.getline(sx,sy+y,a,buf);
                             changebuff(buf,a);
                             self.Putline(dx,dy+y,a,buf);
                           end;
                  32 : for y:=0 to b-1 do
                           begin
                             self.getline(dx,dy+y,a,buf);
                             src^.getline(sx,sy+y,a,buf);
                             changebuff(buf,a);
                             self.Putline(dx,dy+y,a,buf);
                           end;
                end;
                freemem(buf,a*3)
              end;

begin
end.